//-------------------------------------------------------------------------
/*
Copyright (C) 2005 - Justin Marshall

This file is part of the blood reconstruction project.

WinBlood is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

PORTIONS OF CODE FROM DUKE NUKEM 3D SOURCE CODE
*/
//-------------------------------------------------------------------------
// game.h

// Globally disable these warnings
#pragma warning (disable : 4018)
#pragma warning (disable : 4312)

#include <windows.h>
#include <vector>
#include <string>
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
//#include <fmod.h> 
#include <math.h>
#include <time.h>
#include <assert.h>

#include "types.h"

// Temperary fix
#ifndef FORCE_STRLEN
#undef strlen
long strlen(char *intxt);
#endif

//#define FAFTEST // build with shadow warriors FAF code

struct newPlayerInfo_t
{
	long x;
	long y;
	long z;
	long sector;
	long ang;
};

using namespace std;

#define BLOODSTATE_UNKNOWN  0 // for whatever reason state is invalid
#define BLOODSTATE_LOADING  1
#define BLOODSTATE_GAME     2
#define BLOODSTATE_MENU     3

#define PLAYER_WALKSPEED     5
#define PLAYER_RUNSPEED     10

#define PLAYER_WEAPON_PITCHFORK 0
#define PLAYER_WEAPON_FLAREGUN  1
#define PLAYER_WEAPON_STAFF     8

//
// AI
//

#define AI_UNKNOWN    0
#define AI_ZOMBIEMOVE 1

// From the Duke3d Source code
#define NORMALTURN   15	
#define PREAMBLETURN 5  	
#define NORMALKEYMOVE 40 	
#define MAXVEL       ((NORMALKEYMOVE*2)+10)	
#define MAXSVEL      ((NORMALKEYMOVE*2)+10)	
#define MAXANGVEL    127

// Doors
#define LEVEL_ZDOOR      600
#define LEVEL_SLIDEDOOR  614
#define LEVEL_DOORROTATE 617

struct bloodWeapon_t
{
	string weaponName;
	int    weaponID;
	int    ammo;
	int    idlePic;
	bloodWeapon_t::bloodWeapon_t(string weaponName, int weaponID, int ammo,
								 int idlePic)
	{
		this->weaponName = weaponName;
		this->weaponID   = weaponID;
		this->ammo       = ammo;
		this->idlePic    = idlePic;
	}
};

typedef char* buffer32;

struct bloodPlayer_t
{
	int  health;
	int  armor;

	string name;

	long   posx;
	long   posy;
	long   posz;

	long   oldx;
	long   oldy;
	long   oldz;

	long   avel;

	long   posxv;
	long   posxy;

	long   svel;
	long   vel;

	short  ang;
	short  fakangle;

	short  sectorNum;
	short  oldsectornum;
	int  serverID;   // id on the server, given at Login time.  Used only for player 0

	int  thisWeapon;

	vector<bloodWeapon_t *> playerweapons;
};

#define SOSEFFECT_UP   11
#define SOSEFFECT_DOWN 12

// Includes
#include "fileManager.h"
extern fileManager_t   *fileManager;

#include "threads.h"

struct monster_t
{
	int      startPic;
	int      endPic;
	thread_t *mobAI;
};

struct monsterEnt_t
{
	int       mobID;
	int       aiNum;
	monster_t *thisMob;
};

typedef struct
{
	long x, y, z;
	short cstat, picnum;
	signed char shade;
	unsigned char pal, clipdist, filler;
	unsigned char xrepeat, yrepeat;
	signed char xoffset, yoffset;
	short sectnum, statnum;
	short ang, owner, xvel, yvel, zvel;
	short type;
	unsigned short flags;
	short extra;
} spritetype;

//
// xspritetype
// 

typedef struct
{
	byte  unknown1[16]; // fix me
	short data1;
	short data2;
	short data3;
	byte  unknown2[34]; // fix me
} xspritetype;

//
// xsectortype
//

typedef struct
{
	byte unkown1[9];
	byte rxid;
    byte unkown2[34];
	unsigned short marker1;
	unsigned short marker2;
    byte unkown3[14];
} xsectortype;

// 
// xwalltype
// 

typedef struct
{
	byte wallid;
	byte unknown1[5];
	byte rxid;
	byte unknown2[17];
} xwalltype;

typedef struct
{
	short wallptr, wallnum;
	long ceilingz, floorz;
	short ceilingstat, floorstat;
	short ceilingpicnum, ceilingheinum;
	signed char ceilingshade;
	unsigned char ceilingpal, ceilingxpanning, ceilingypanning;
	short floorpicnum, floorheinum;
	signed char floorshade;
	unsigned char floorpal, floorxpanning, floorypanning;
	unsigned char visibility, filler;
	short lotag, hitag, extra;
} sectortype;

/* 32 bytes */
typedef struct
{
	long x, y;
	short point2, nextwall, nextsector, cstat;
	short picnum, overpicnum;
	signed char shade;
	char pal, xrepeat, yrepeat, xpanning, ypanning;
	short lotag, hitag, extra;
} walltype;

struct doorSlideEvent_t
{
	long sectornum;
	long openx;
	long openy;
	long openz;

	long closex;
	long closey;
	long closez;

	float starttic;
	float endtic;

	short beginopen;
	short endopen;
	short endclose;
	short startclose;

	doorSlideEvent_t::doorSlideEvent_t(long openx,  long openy, long openz,
									   long closex, long closey, long closez,
									   long sectornum)
	{
		this->openx     = openx;
		this->openy     = openy;
		this->openz     = openz;
		this->closex    = closex;
		this->closey    = closey;
		this->closez    = closez;
		this->sectornum = sectornum;

		// This isn't determined here..
		this->starttic = 0;
		this->endtic   = 0;

		this->beginopen  = 0;
		this->endopen    = 0;
		this->endclose   = 0;
		this->startclose = 0;
	}
};

struct worldEvent_t
{
	int  type;
	int  flagged;
	int  switch_t;
	void *data;
};

struct bloodsfx_t
{
	char   soundName[20];
	int    frequency;
	int    looping;
	int    volume;
};

struct door_t
{
	long minX;
	long minY;
	
	long maxX;
	long maxY;

	long startZ;
	long endZ;
	int  sectNum;

	int openTime;

	door_t::door_t(long minx, long miny, long maxX, long maxY, long startZ,
				   long endZ, int sectNum)
	{
		this->minX = minx;
		this->maxY = miny;
		
		this->maxX = maxX;
		this->maxY = maxY;

		openTime   = 0; // doors inital state is closed

		this->startZ  = startZ;
		this->endZ    = endZ;
		this->sectNum = sectNum;
	}
};

struct jprojectTile
{
	int damage;
	int spritenum;
	int lastpic;
	int explosion;
};

struct gameCFG_t
{
	int fullscreen;
	int renderer;
	int resolution;
	int antistrophy; 
};

struct gameSprites_t
{
	int spriteNum;
	int health;
	int pain;
};

struct hiddensprites
{
	int starttile;
	int endtile;
	int type;
};

struct tsectortype
{
	long slope;
	long sectnum;
	long pic;
	long zval;
};

struct AIIMAGES
{
	int type;
	int start;
	int end;
};

/*
	jCommon

	Functions that deal directlly with Build.
*/

class jCommon
{
	public:
		int spriteSeePlayer(int spriteNum);
		boolean PicInView(short tile_num, BOOL reset);
};
#include "ai.h"

/*
	Class game
	Game system, uses all underlying functions to run this game.
*/

class Game
{
	friend class soundManager_t;
	friend class sectorManager_t;

	public:
		Game();
		int     RunFrame(int totalclock);
		void    LoadBoard(char *board);
		boolean g_PlayerInGame(void);
		void    DisplayLevel(void);
		void    Error(char *err, ...);
		void    Shutdown(void);
		void    Register_Weapon(char *weaponScript);
		void    RegisterAI(int startPic, int endPic, char *script_name);
		void    ModifySprite(int spriteNum, spritetype *thisSprite);
		void    StartAI(void);
		void    ShutdownAI(void);
		void    SuspendAI(void);
		void	ReRunAI(void);
		void    ShotWeapon(int damage);
		void    PlayerSetLocation(int playerNum);
		void    RunEffects(void);
		void    PauseWeaponThreads(void);
		void    FatalCrash(char *err);
		void    WriteLog(char *thisTxt);
		void    ResumeWeaponThread(void);
		void    SetSpriteStates(void);
		void    RunOperators(void);
		void    PlayerFireWeapon(int projectileTile, int damage, int explosion, int range, int deathspawn);
		void    SectorFOF(long x,long y,long z,long a,long h, long smoothratio);
		void    FOFDraw(int spnum,long x,long y,long z,short a,short h,long smoothratio);
		void    EmulateFOF(int spriteNum,long x,long y,long z,long a,long h, long smoothratio);
		void    CheckExpansions(void);
		void    DrawHudNumber(int number, long startfont, int shade, int x, int y);
		void    MoveSector(sectortype *thisSector, long startx, long starty, long startz, long oldang,long newx, long newy, long newz, long newang, short rxid);
		void    entWonder(int spriteNum, int tics, int frontWalk, int frontWalk2);
		void    DrawAutoMap(void);
		void    WriteBloodCFG(gameCFG_t *thisCfg);
		void    ReadBloodCFG(void);
		void    FOFDraw(long x,long y,long z,int spnum, long smoothratio);
		void    MoveSectorXYZ(long sectornum, long x, long y, long z);
		void    LoadVoxels(void);
		boolean MobInMeleeRange(int sprNum);
		void    MobAttackPlayer(int damage);
		void    MobExecuteAI(int aiNum, int spritenum);
		long    MoveSprite(spritetype *thisSprite, long fvel, long svel, unsigned long cliptype);
		void    makeitfall(int spriteNum);
		void    PrecacheSounds(void);
		void    PlaySfxSprite(short soundid);
		void    RotatePoint(long *x, long *y, long angVect2, long startx, long starty, long pivx, long pivy);
		void    WorldAmbient(void);
		void    InitExtras(void);
		void    OperateSprites(void);
		void    RunProjectiles(void);
		void    RunSpecials(void);
		void    SpriteAnimation(int spritenum, int startTile, int frames, int time, int x, int y);
		void    OperateSpritesAuto(void);
		void    DeletePicALL(long picnum);
		void    PushSectors(void);
		void    PopSectors(void);
		void    updatesectorz(long x, long y, long z, short *sectnum);
		void    MobMain(int spritenum);
		boolean mobCheckObstacles(spritetype *thisSprite, int ang, int dist);
		int     mobGetAIType(int picnum);
		void    InternalRegisterAI(void);
		boolean mobCheckMove(spritetype *thisSprite, int dist);
		int     mobChangePic(int baseFrame1, int baseFrame2, int baseFrame3,int baseFrame4,
					   int cycle, int framenum, int maxframes);

		//  SHADOW WARRIOR FAF CODE
		void    FAFDraw(long tx, long ty, long tz, short tang, long thoriz, short tsectnum);
		void    DeletePic(short sectnum, short floor);
		short   ViewSectorInScene(short cursectnum, short type, short type2);
		boolean FindCeilingView(short match, LONGp x, LONGp y, LONG z, SHORTp sectnum);
		void    GetUpperLowerSector(short match, long x, long y, short *upper, short *lower);
	    boolean FindFloorView(short match, LONGp x, LONGp y, LONG z, SHORTp sectnum);

		int     forcedProjWidth;
		int     forcedProjHeight;

		boolean NodrawWeapon;
		boolean autoMap;
		boolean noclip;
		int     brightness;
		int     levelMusic;

		int     thisWeaponPic;
		int     thisWeaponY;
		int     thisWeaponX;
		int     weaponInUse;

		vector<bloodPlayer_t *> gamePlayers;
		vector<long> cachedCeilingZ;
		vector<long> cachedFloorZ;
		int     expansionLists;
		float   jumpingtics;
		int     jumpturn;
		// Some weapons have two parts to them
		int     thisWeaponPic2;
		int     thisWeaponY2;
		int     thisWeaponX2;
		int     weaponInUse2;
		int     skippageflip;

		boolean multiplayer;
		boolean consoleactive;

		void    SetPlayerWeapon(int weaponNum);

		int     SetWeaponPic(int picNum)
		{
			this->thisWeaponPic = picNum;
		}

		static aiConstruct activeMonsters[];
		vector<int>             deadSprites;
		vector<thread_t *>      playerWeapons;
		vector<monster_t *>     registeredAI;
		vector<monsterEnt_t *>  worldMobs;
		vector<worldEvent_t *>  worldEvents;
		vector<gameSprites_t *> gameSprites;
		vector<char *>			gameAIScripts;
		vector<jprojectTile *>  activeProjectiles;
		vector<jprojectTile *>  activeSpecials;
		vector<spritetype  *>   flatsprites;
		vector<tsectortype>     saveSectors;
		boolean scripts_loaded;

		boolean exit_game;

		thread_t *gameThread;

		void LoadingPlague(void);
		void EndLoadingPlague(void);

		char *thisBoard;

		boolean loadingplague;

		vector<int> removeSprites;

		int thisRenderer;
		int anisotropy;
		int fullscreen;
		int resolution;
		int numSectors;
		int numofsprites;
	private:
		void    OperateSectors(void);
		void    init_system(void);
		void    DrawHUD(void);
		void    P_HandleMovement(void);
		void    Level_MakeActors(void);
		void    ThreadTest(void);
		void    RemoveEffectors(void);
		void    ShutdownWeaponThreads(void);
		void    KillAllThreads(void);
		void    CheckArgs(void);
		int initframe;
		boolean ingame;
		static hiddensprites InvisibleSprites[];

		static bloodsfx_t RegisteredSFX[];
		FILE *thislogFile;
	protected:
		int gamestate;
};

#define Z(value) ((long)(value) << 8)
#define PIXZ(value) ((long)(value) >> 8)

#include "guiManager.h"
#include "game_shared.h"
#include "game_menu.h"
#include "keyboard.h"
#include "bloodTiles.h"
#include "soundManager.h"
#include "buildMath.h"
#include "gameProgs.h"
#include "network.h"
#include "console.h"
#include "sectorManager.h"
#include "mathutil.h"
#include "cheats.h"

#ifndef min
#define min(a, b)  (((a) < (b)) ? (a) : (b))
#endif

#ifndef max
#define max(a, b)  (((a) > (b)) ? (a) : (b))
#endif

struct build_func_t
{
	void (*nextpage)(void);
	boolean (*build_KeyDown)(byte key);
	void (*Load_Pallete)(char *filename);
	void (*rotatesprite)(long sx, long sy, long z, short a, short picnum,
                  signed char dashade, char dapalnum, char dastat,
                  long cx1, long cy1, long cx2, long cy2);
	void (*clearview)(long dacol);
	long   (*loadboard)(char *filename, char fromwhere, long *daposx, long *daposy, long *daposz, short *daang, short *dacursectnum);
	void (*SDL_Quit)(void);
	int  (*Decrypt_Map)(char *in, char *out);
	void (*displayLevel)(long posx, long posy, long posz, long ang,short snum,long smoothratio);
	long (*scale)(long i1, long i2, long i3);
	newPlayerInfo_t (*P_Movecode)(short ang, long fvel, long svel);
	int (*clipmove) (long *x, long *y, long *z, short *sectnum,
              long xvect, long yvect, long walldist, long ceildist,
              long flordist, unsigned long cliptype);

	int (*mulscale) (int input1, int input2, int input3);
	int (*pushmove)(long *x, long *y, long *z, short *sectnum,
             long walldist, long ceildist, long flordist,
             unsigned long cliptype);
	int (*ftotalclock)(void);
	void (*Set_Brightness)(int brightness);
	void (*getzrange)(long x, long y, long z, short sectnum,
			long *ceilz, long *ceilhit, long *florz, long *florhit,
			long walldist, unsigned long cliptype);
	spritetype *(*Game_GetSprite)(int i);
	void (*Justin_SetSky)(long pScale, short newpskyoff[256]);
	void (*build_ClearKey)(byte key);
	int (*neartag)(long xs, long ys, long zs, short sectnum, short ange,
            short *neartagsector, short *neartagwall, short *neartagsprite,
            long *neartaghitdist, long neartagrange, char tagsearch);
	sectortype *(*GetSector)(int sectNum);
	int (*hitscan)(long xs, long ys, long zs, short sectnum,
            long vx, long vy, long vz,
	        short *hitsect, short *hitwall, short *hitsprite,
	        long *hitx, long *hity, long *hitz, unsigned long cliptype);
	int (*krand)(void);
	void (*setviewtotile)(short tilenume, long xsiz, long ysiz);
	void (*setviewback)(void);
	short (*getTileX)(int tileNum);
	short (*getTileY)(int tileNum);
	void (*loadtile)(short tilenume);
	short *(*getSinTable)(void);
	short sintable[2048];
	char *(*getParam)(int param);
	walltype *(*get_Wall)(int wallNum);
	int (*cansee)(long x1, long y1, long z1, short sect1,
            long x2, long y2, long z2, short sect2);
	void (*drawline256) (long x1, long y1, long x2, long y2, unsigned char col);
	void (*Change_ScreenInfo)(int width, int height, int bpp, int antistrophy, int glBased, int fullscreen);
	xspritetype *(*GetXSprite)(int spritenum);
	void (*KBClearLastScanCode)(void);
	short (*getincangle)(short a,short na);
	xsectortype *(*GetXSector)(int sectorNum);
	xwalltype *(*GetXWall)(int wallNum);
	void   (*dragpoint)(short pointhighlight, long dax, long day);
	void   (*rotatepoint)(long xpivot, long ypivot, long x, long y, short daang, long *x2, long *y2);
	void   (*InsertSprite)(short sectnum, short sprnum, long x, long y, long z);
	long   (*spawnsprite)(long newspriteindex2, long x2, long y2, long z2, short cstat2, char shade2, char pal2,         
		                       char clipdist2, unsigned char xrepeat2, unsigned char yrepeat2, char xoffset2, char yoffset2, short picnum2, short ang2,       
		                       short xvel2, short yvel2, short zvel2, short owner2, long sectnum2, long statnum2, short lotag2, short hitag2, long extra2);
	long   (*deletesprite)(short spritenum);
	spritetype *(*unsafeGetSprite)(int i);
	long   (*changespritesect)(short spritenum, short newsectnum);
	boolean (*playerTouchSprite)(long playerx, long playery, long playerz,
						         long psectornum, long picnum);
	void   (*updatesector)(long x, long y, short *sectnum);
	char   (*build_gotpic)(long picnum);
	void   (*build_resetgotpic)(long picnum);
	long   (*inside)(long x, long y, short sectnum);
	long   (*getangle)(long xvect, long yvect);
};

#define TICSPERFRAME 3
#define CLIPMASK0 (((1L)<<16)+1L)
#define CLIPMASK1 (((256L)<<16)+64L)

// Game Managers
extern Game            *thisGame;
extern guiManager      *guimanager;
extern fileManager_t   *fileManager;
extern gameMenu_t      *menuManager;
extern soundManager_t  *soundManager;
extern build_func_t    *build_func;
extern kenNet_t        *netManager;
extern console_t       *console;
extern sectorManager_t *sectorManager;
extern jCommon         *common;